;; Generated package description from nano-agenda.el  -*- no-byte-compile: t -*-
(define-package "nano-agenda" "0.2.1" "N Λ N O agenda" '((emacs "27.1")) :maintainer '("Nicolas P. Rougier" . "Nicolas.Rougier@inria.fr") :keywords '("convenience" "org-mode" "org-agenda") :url "https://github.com/rougier/nano-agenda")
